@@ -8,6 +8,7 @@
 //! This module implements operations for a simple asynchronized mysql database
 
 use akd::errors::StorageError;
+use akd::history_tree_node::{HistoryTreeNode, NodeKey};
 use akd::node_state::NodeLabel;
 use akd::storage::types::{
     AkdKey, DbRecord, KeyData, StorageType, ValueState, ValueStateRetrievalFlag,
@@ -39,9 +40,9 @@ const TEMP_IDS_TABLE: &str = "temp_ids_table";
 const MAXIMUM_SQL_TIER_CONNECTION_TIMEOUT_SECS: u64 = 300;
 const SQL_RECONNECTION_DELAY_SECS: u64 = 5;
 
-const SELECT_AZKS_DATA: &str = "`root`, `epoch`, `num_nodes`";
+const SELECT_AZKS_DATA: &str = "`epoch`, `num_nodes`";
 const SELECT_HISTORY_TREE_NODE_DATA: &str =
-    "`location`, `label_len`, `label_val`, `epochs`, `parent`, `node_type`";
+    "`label_len`, `label_val`, `birth_epoch`, `last_epoch`, `parent_label_len`, `parent_label_val`, `node_type`";
 const SELECT_HISTORY_NODE_STATE_DATA: &str =
     "`label_len`, `label_val`, `epoch`, `value`, `child_states`";
 const SELECT_USER_DATA: &str =
@@ -329,18 +330,18 @@ impl<'a> AsyncMySqlDatabase {
         // AZKS table
         let command = "CREATE TABLE IF NOT EXISTS `".to_owned()
             + TABLE_AZKS
-            + "` (`key` SMALLINT UNSIGNED NOT NULL, `root` BIGINT UNSIGNED NOT NULL,"
-            + " `epoch` BIGINT UNSIGNED NOT NULL, `num_nodes` BIGINT UNSIGNED NOT NULL,"
-            + " PRIMARY KEY (`key`))";
+            + "` (`key` SMALLINT UNSIGNED NOT NULL, `epoch` BIGINT UNSIGNED NOT NULL,"
+            + " `num_nodes` BIGINT UNSIGNED NOT NULL, PRIMARY KEY (`key`))";
         tx.query_drop(command).await?;
 
         // History tree nodes table
         let command = "CREATE TABLE IF NOT EXISTS `".to_owned()
             + TABLE_HISTORY_TREE_NODES
-            + "` (`location` BIGINT UNSIGNED NOT NULL, `label_len` INT UNSIGNED NOT NULL,"
-            + " `label_val` BIGINT UNSIGNED NOT NULL, `epochs` VARBINARY(2000),"
-            + " `parent` BIGINT UNSIGNED NOT NULL, `node_type` SMALLINT UNSIGNED NOT NULL,"
-            + " PRIMARY KEY (`location`))";
+            + "` (`label_len` INT UNSIGNED NOT NULL, `label_val` BIGINT UNSIGNED NOT NULL,"
+            + "  `birth_epoch` BIGINT UNSIGNED NOT NULL,"
+            + " `last_epoch` BIGINT UNSIGNED NOT NULL, `parent_label_len` INT UNSIGNED NOT NULL,"
+            + " `parent_label_val` BIGINT UNSIGNED NOT NULL, `node_type` SMALLINT UNSIGNED NOT NULL,"
+            + " PRIMARY KEY (`label_len`, `label_val`))";
         tx.query_drop(command).await?;
 
         // History node states table
@@ -386,6 +387,28 @@ impl<'a> AsyncMySqlDatabase {
         Ok(())
     }
 
+    /// Drop all the tables
+    pub async fn drop_tables(&self) -> core::result::Result<(), MySqlError> {
+        let mut conn = self.get_connection().await?;
+        let mut tx = conn.start_transaction(TxOpts::default()).await?;
+
+        let command = "DROP TABLE IF EXISTS `".to_owned() + TABLE_AZKS + "`";
+        tx.query_drop(command).await?;
+
+        let command = "DROP TABLE IF EXISTS `".to_owned() + TABLE_USER + "`";
+        tx.query_drop(command).await?;
+
+        let command = "DROP TABLE IF EXISTS `".to_owned() + TABLE_HISTORY_NODE_STATES + "`";
+        tx.query_drop(command).await?;
+
+        let command = "DROP TABLE IF EXISTS `".to_owned() + TABLE_HISTORY_TREE_NODES + "`";
+        tx.query_drop(command).await?;
+
+        tx.commit().await?;
+
+        Ok(())
+    }
+
     /// Storage a record in the data layer
     async fn internal_set(
         &self,
@@ -456,9 +479,7 @@ impl<'a> AsyncMySqlDatabase {
                 DbRecord::HistoryNodeState(_) => {
                     DbRecord::set_batch_statement::<akd::node_state::HistoryNodeState>(i)
                 }
-                DbRecord::HistoryTreeNode(_) => {
-                    DbRecord::set_batch_statement::<akd::history_tree_node::HistoryTreeNode>(i)
-                }
+                DbRecord::HistoryTreeNode(_) => DbRecord::set_batch_statement::<HistoryTreeNode>(i),
                 DbRecord::ValueState(_) => {
                     DbRecord::set_batch_statement::<akd::storage::types::ValueState>(i)
                 }
@@ -521,29 +542,6 @@ impl<'a> AsyncMySqlDatabase {
         Ok(())
     }
 
-    /// Cleanup the test data table
-    #[allow(dead_code)]
-    pub async fn test_cleanup(&self) -> core::result::Result<(), MySqlError> {
-        let mut conn = self.get_connection().await?;
-        let mut tx = conn.start_transaction(TxOpts::default()).await?;
-
-        let command = "DROP TABLE IF EXISTS `".to_owned() + TABLE_AZKS + "`";
-        tx.query_drop(command).await?;
-
-        let command = "DROP TABLE IF EXISTS `".to_owned() + TABLE_USER + "`";
-        tx.query_drop(command).await?;
-
-        let command = "DROP TABLE IF EXISTS `".to_owned() + TABLE_HISTORY_NODE_STATES + "`";
-        tx.query_drop(command).await?;
-
-        let command = "DROP TABLE IF EXISTS `".to_owned() + TABLE_HISTORY_TREE_NODES + "`";
-        tx.query_drop(command).await?;
-
-        tx.commit().await?;
-
-        Ok(())
-    }
-
     fn try_dockers() -> std::io::Result<std::process::Output> {
         let potential_docker_paths = vec![
             "/usr/local/bin/docker",
@@ -621,7 +619,7 @@ impl Storage for AsyncMySqlDatabase {
         let mut node_state_size = "Node state count: Query err".to_string();
         let mut value_state_size = "Value state count: Query err".to_string();
         if let Ok(mut conn) = self.get_connection().await {
-            let query_text = format!("SELECT COUNT(`location`) FROM {}", TABLE_HISTORY_TREE_NODES);
+            let query_text = format!("SELECT COUNT(*) FROM {}", TABLE_HISTORY_TREE_NODES);
             if let Ok(results) = conn.query_iter(query_text).await {
                 if let Ok(mapped) = results
                     .map_and_drop(|row| {
@@ -819,7 +817,7 @@ impl Storage for AsyncMySqlDatabase {
                         }
                         DbRecord::HistoryTreeNode(node) => {
                             if let DbRecord::HistoryTreeNode(node2) = &b {
-                                node.location.cmp(&node2.location)
+                                node.label.cmp(&node2.label)
                             } else {
                                 Ordering::Equal
                             }
@@ -1437,6 +1435,52 @@ impl Storage for AsyncMySqlDatabase {
             Err(code) => Err(StorageError::GetError(code.to_string())),
         }
     }
+
+    async fn get_epoch_lte_epoch(
+        &self,
+        node_label: akd::node_state::NodeLabel,
+        epoch_in_question: u64,
+    ) -> core::result::Result<u64, StorageError> {
+        *(self.num_reads.write().await) += 1;
+
+        let result = async {
+            let tic = Instant::now();
+
+            let mut conn = self.get_connection().await?;
+
+            let statement = format!("SELECT `epoch` FROM {} WHERE `label_len` = :len AND `label_val` = :val AND `epoch` <= :epoch ORDER BY `epoch` DESC LIMIT 1", TABLE_HISTORY_NODE_STATES);
+            let out = conn
+                .exec_first(
+                    statement,
+                    params! {
+                        "len" => node_label.len,
+                        "val" => node_label.val,
+                        "epoch" => epoch_in_question,
+                    },
+                )
+                .await;
+
+            let toc = Instant::now() - tic;
+            *(self.time_read.write().await) += toc;
+
+            let result = self.check_for_infra_error(out)?;
+
+            match result {
+                Some(r) => Ok::<_, MySqlError>(r),
+                None => Ok::<_, MySqlError>(u64::MAX),
+            }
+        };
+
+        debug!("END MySQL get epoch LTE epoch");
+        match result.await {
+            Ok(u64::MAX) => Err(StorageError::GetError(format!(
+                "Node (val: {}, len: {}) did not exist <= epoch {}",
+                node_label.val, node_label.len, epoch_in_question
+            ))),
+            Ok(ep) => Ok(ep),
+            Err(code) => Err(StorageError::GetError(code.to_string())),
+        }
+    }
 }
 
 /* Generic data structure handling for MySQL */
@@ -1468,17 +1512,14 @@ trait MySqlStorable {
     fn from_row<St: Storable>(row: &mut mysql_async::Row) -> core::result::Result<Self, MySqlError>
     where
         Self: std::marker::Sized;
-
-    fn serialize_epochs(epochs: &[u64]) -> Vec<u8>;
-    fn deserialize_epochs(bin: &[u8]) -> Option<Vec<u64>>;
 }
 
 impl MySqlStorable for DbRecord {
     fn set_statement(&self) -> String {
         match &self {
-            DbRecord::Azks(_) => format!("INSERT INTO `{}` (`key`, {}) VALUES (:key, :root, :epoch, :num_nodes) ON DUPLICATE KEY UPDATE `root` = :root, `epoch` = :epoch, `num_nodes` = :num_nodes", TABLE_AZKS, SELECT_AZKS_DATA),
+            DbRecord::Azks(_) => format!("INSERT INTO `{}` (`key`, {}) VALUES (:key, :epoch, :num_nodes) ON DUPLICATE KEY UPDATE `epoch` = :epoch, `num_nodes` = :num_nodes", TABLE_AZKS, SELECT_AZKS_DATA),
             DbRecord::HistoryNodeState(_) => format!("INSERT INTO `{}` ({}) VALUES (:label_len, :label_val, :epoch, :value, :child_states) ON DUPLICATE KEY UPDATE `value` = :value, `child_states` = :child_states", TABLE_HISTORY_NODE_STATES, SELECT_HISTORY_NODE_STATE_DATA),
-            DbRecord::HistoryTreeNode(_) => format!("INSERT INTO `{}` ({}) VALUES (:location, :label_len, :label_val, :epochs, :parent, :node_type) ON DUPLICATE KEY UPDATE `label_len` = :label_len, `label_val` = :label_val, `epochs` = :epochs, `parent` = :parent, `node_type` = :node_type", TABLE_HISTORY_TREE_NODES, SELECT_HISTORY_TREE_NODE_DATA),
+            DbRecord::HistoryTreeNode(_) => format!("INSERT INTO `{}` ({}) VALUES (:label_len, :label_val, :birth_epoch, :last_epoch, :parent_label_len, :parent_label_val, :node_type) ON DUPLICATE KEY UPDATE `label_len` = :label_len, `label_val` = :label_val, `birth_epoch` = :birth_epoch, `last_epoch` = :last_epoch, `parent_label_len` = :parent_label_len, `parent_label_val` = :parent_label_val, `node_type` = :node_type", TABLE_HISTORY_TREE_NODES, SELECT_HISTORY_TREE_NODE_DATA),
             DbRecord::ValueState(_) => format!("INSERT INTO `{}` ({}) VALUES (:username, :epoch, :version, :node_label_val, :node_label_len, :data)", TABLE_USER, SELECT_USER_DATA),
         }
     }
@@ -1486,7 +1527,7 @@ impl MySqlStorable for DbRecord {
     fn set_params(&self) -> mysql_async::Params {
         match &self {
             DbRecord::Azks(azks) => {
-                params! { "key" => 1u8, "root" => azks.root, "epoch" => azks.latest_epoch, "num_nodes" => azks.num_nodes }
+                params! { "key" => 1u8, "epoch" => azks.latest_epoch, "num_nodes" => azks.num_nodes }
             }
             DbRecord::HistoryNodeState(state) => {
                 let bin_data = bincode::serialize(&state.child_states).unwrap();
@@ -1494,8 +1535,7 @@ impl MySqlStorable for DbRecord {
                 params! { "label_len" => id.0.len, "label_val" => id.0.val, "epoch" => id.1, "value" => state.value.clone(), "child_states" => bin_data }
             }
             DbRecord::HistoryTreeNode(node) => {
-                let bin_data = DbRecord::serialize_epochs(&node.epochs);
-                params! { "location" => node.location, "label_len" => node.label.len, "label_val" => node.label.val, "epochs" => bin_data, "parent" => node.parent, "node_type" => node.node_type as u8 }
+                params! { "label_len" => node.label.len, "label_val" => node.label.val, "birth_epoch" => node.birth_epoch, "last_epoch" => node.last_epoch, "parent_label_len" => node.parent.len, "parent_label_val" => node.parent.val, "node_type" => node.node_type as u8 }
             }
             DbRecord::ValueState(state) => {
                 params! { "username" => state.get_id().0, "epoch" => state.epoch, "version" => state.version, "node_label_len" => state.label.len, "node_label_val" => state.label.val, "data" => state.plaintext_val.0.clone() }
@@ -1514,10 +1554,16 @@ impl MySqlStorable for DbRecord {
                     );
                 }
                 StorageType::HistoryTreeNode => {
-                    parts = format!("{}(:location{}, :label_len{}, :label_val{}, :epochs{}, :parent{}, :node_type{})", parts, i, i, i, i, i, i);
+                    parts = format!(
+                        "{}(:label_len{}, :label_val{}, :birth_epoch{}, :last_epoch{}, :parent_label_len{}, :parent_label_val{}, :node_type{})",
+                        parts, i, i, i, i, i, i, i
+                    );
                 }
                 StorageType::ValueState => {
-                    parts = format!("{}(:username{}, :epoch{}, :version{}, :node_label_val{}, :node_label_len{}, :data{})", parts, i, i, i, i, i, i);
+                    parts = format!(
+                        "{}(:username{}, :epoch{}, :version{}, :node_label_val{}, :node_label_len{}, :data{})",
+                        parts, i, i, i, i, i, i
+                    );
                 }
                 _ => {
                     // azks
@@ -1530,9 +1576,9 @@ impl MySqlStorable for DbRecord {
         }
 
         match St::data_type() {
-            StorageType::Azks => format!("INSERT INTO `{}` (`key`, {}) VALUES (:key, :root, :epoch, :num_nodes) as new ON DUPLICATE KEY UPDATE `root` = new.root, `epoch` = new.epoch, `num_nodes` = new.num_nodes", TABLE_AZKS, SELECT_AZKS_DATA),
+            StorageType::Azks => format!("INSERT INTO `{}` (`key`, {}) VALUES (:key, :epoch, :num_nodes) as new ON DUPLICATE KEY UPDATE `epoch` = new.epoch, `num_nodes` = new.num_nodes", TABLE_AZKS, SELECT_AZKS_DATA),
             StorageType::HistoryNodeState => format!("INSERT INTO `{}` ({}) VALUES {} as new ON DUPLICATE KEY UPDATE `value` = new.value, `child_states` = new.child_states", TABLE_HISTORY_NODE_STATES, SELECT_HISTORY_NODE_STATE_DATA, parts),
-            StorageType::HistoryTreeNode => format!("INSERT INTO `{}` ({}) VALUES {} as new ON DUPLICATE KEY UPDATE `label_len` = new.label_len, `label_val` = new.label_val, `epochs` = new.epochs, `parent` = new.parent, `node_type` = new.node_type", TABLE_HISTORY_TREE_NODES, SELECT_HISTORY_TREE_NODE_DATA, parts),
+            StorageType::HistoryTreeNode => format!("INSERT INTO `{}` ({}) VALUES {} as new ON DUPLICATE KEY UPDATE `label_len` = new.label_len, `label_val` = new.label_val, `birth_epoch` = new.birth_epoch, `last_epoch` = new.last_epoch, `parent_label_len` = new.parent_label_len, `parent_label_val` = new.parent_label_val, `node_type` = new.node_type", TABLE_HISTORY_TREE_NODES, SELECT_HISTORY_TREE_NODE_DATA, parts),
             StorageType::ValueState => format!("INSERT INTO `{}` ({}) VALUES {}", TABLE_USER, SELECT_USER_DATA, parts),
         }
     }
@@ -1545,7 +1591,6 @@ impl MySqlStorable for DbRecord {
                 DbRecord::Azks(azks) => {
                     vec![
                         ("key".to_string(), Value::from(1u8)),
-                        ("root".to_string(), Value::from(azks.root)),
                         ("epoch".to_string(), Value::from(azks.latest_epoch)),
                         ("num_nodes".to_string(), Value::from(azks.num_nodes)),
                     ]
@@ -1562,13 +1607,19 @@ impl MySqlStorable for DbRecord {
                     ]
                 }
                 DbRecord::HistoryTreeNode(node) => {
-                    let bin_data = DbRecord::serialize_epochs(&node.epochs);
                     vec![
-                        (format!("location{}", idx), Value::from(node.location)),
                         (format!("label_len{}", idx), Value::from(node.label.len)),
                         (format!("label_val{}", idx), Value::from(node.label.val)),
-                        (format!("epochs{}", idx), Value::from(bin_data)),
-                        (format!("parent{}", idx), Value::from(node.parent)),
+                        (format!("birth_epoch{}", idx), Value::from(node.birth_epoch)),
+                        (format!("last_epoch{}", idx), Value::from(node.last_epoch)),
+                        (
+                            format!("parent_label_len{}", idx),
+                            Value::from(node.parent.len),
+                        ),
+                        (
+                            format!("parent_label_val{}", idx),
+                            Value::from(node.parent.val),
+                        ),
                         (
                             format!("node_type{}", idx),
                             Value::from(node.node_type as u8),
@@ -1630,7 +1681,7 @@ impl MySqlStorable for DbRecord {
             StorageType::HistoryTreeNode => {
                 Some(
                     format!(
-                        "CREATE TEMPORARY TABLE `{}`(`location` BIGINT UNSIGNED NOT NULL, PRIMARY KEY(`location`))",
+                        "CREATE TEMPORARY TABLE `{}`(`label_len` INT UNSIGNED NOT NULL, `label_val` BIGINT UNSIGNED NOT NULL, PRIMARY KEY(`label_len`, `label_val`))",
                         TEMP_IDS_TABLE
                     )
                 )
@@ -1656,7 +1707,10 @@ impl MySqlStorable for DbRecord {
                 )
             }
             StorageType::HistoryTreeNode => {
-                format!("INSERT INTO `{}` (`location`) VALUES ", TEMP_IDS_TABLE)
+                format!(
+                    "INSERT INTO `{}` (`label_len`, `label_val`) VALUES ",
+                    TEMP_IDS_TABLE
+                )
             }
             StorageType::ValueState => {
                 format!(
@@ -1673,7 +1727,7 @@ impl MySqlStorable for DbRecord {
                         format!("(:label_len{}, :label_val{}, :epoch{})", i, i, i)
                     }
                     StorageType::HistoryTreeNode => {
-                        format!("(:location{})", i)
+                        format!("(:label_len{}, :label_val{})", i, i)
                     }
                     StorageType::ValueState => {
                         format!("(:username{}, :epoch{})", i, i)
@@ -1690,7 +1744,7 @@ impl MySqlStorable for DbRecord {
             statement += match St::data_type() {
                 StorageType::Azks => "",
                 StorageType::HistoryNodeState => "(:label_len, :label_val, :epoch)",
-                StorageType::HistoryTreeNode => "(:location)",
+                StorageType::HistoryTreeNode => "(:label_len, :label_val)",
                 StorageType::ValueState => "(:username, :epoch)",
             };
         }
@@ -1711,7 +1765,7 @@ impl MySqlStorable for DbRecord {
             }
             StorageType::HistoryTreeNode => {
                 format!(
-                    "SELECT a.`location`, a.`label_len`, a.`label_val`, a.`epochs`, a.`parent`, a.`node_type` FROM `{}` a INNER JOIN {} ids ON ids.`location` = a.`location`",
+                    "SELECT a.`label_len`, a.`label_val`, a.`birth_epoch`, a.`last_epoch`, a.`parent_label_len`, a.`parent_label_val`, a.`node_type` FROM `{}` a INNER JOIN {} ids ON ids.`label_len` = a.`label_len` AND ids.`label_val` = a.`label_val`",
                     TABLE_HISTORY_TREE_NODES,
                     TEMP_IDS_TABLE
                 )
@@ -1730,7 +1784,7 @@ impl MySqlStorable for DbRecord {
         match St::data_type() {
             StorageType::Azks => format!("SELECT {} FROM `{}` LIMIT 1", SELECT_AZKS_DATA, TABLE_AZKS),
             StorageType::HistoryNodeState => format!("SELECT {} FROM `{}` WHERE `label_len` = :label_len AND `label_val` = :label_val AND `epoch` = :epoch", SELECT_HISTORY_NODE_STATE_DATA, TABLE_HISTORY_NODE_STATES),
-            StorageType::HistoryTreeNode => format!("SELECT {} FROM `{}` WHERE `location` = :location", SELECT_HISTORY_TREE_NODE_DATA, TABLE_HISTORY_TREE_NODES),
+            StorageType::HistoryTreeNode => format!("SELECT {} FROM `{}` WHERE `label_len` = :label_len AND `label_val` = :label_val", SELECT_HISTORY_TREE_NODE_DATA, TABLE_HISTORY_TREE_NODES),
             StorageType::ValueState => format!("SELECT {} FROM `{}` WHERE `username` = :username AND `epoch` = :epoch", SELECT_USER_DATA, TABLE_USER),
         }
     }
@@ -1765,11 +1819,14 @@ impl MySqlStorable for DbRecord {
                     .enumerate()
                     .map(|(idx, key)| {
                         let bin = St::get_full_binary_key_id(key);
-                        let back: akd::history_tree_node::NodeKey =
-                            akd::history_tree_node::HistoryTreeNode::key_from_full_binary(&bin)
-                                .unwrap();
-                        (format!("location{}", idx), Value::from(back.0))
+                        let back: NodeKey = HistoryTreeNode::key_from_full_binary(&bin).unwrap();
+                        vec![
+                            (format!("label_len{}", idx), Value::from(back.0.len)),
+                            (format!("label_val{}", idx), Value::from(back.0.val)),
+                        ]
                     })
+                    .into_iter()
+                    .flatten()
                     .collect::<Vec<_>>();
                 Some(mysql_async::Params::from(pvec))
             }
@@ -1809,10 +1866,10 @@ impl MySqlStorable for DbRecord {
             }
             StorageType::HistoryTreeNode => {
                 let bin = St::get_full_binary_key_id(key);
-                let back: akd::history_tree_node::NodeKey =
-                    akd::history_tree_node::HistoryTreeNode::key_from_full_binary(&bin).unwrap();
+                let back: NodeKey = HistoryTreeNode::key_from_full_binary(&bin).unwrap();
                 Some(params! {
-                    "location" => back.0
+                    "label_len" => back.0.len,
+                    "label_val" => back.0.val,
                 })
             }
             StorageType::ValueState => {
@@ -1833,11 +1890,9 @@ impl MySqlStorable for DbRecord {
     {
         match St::data_type() {
             StorageType::Azks => {
-                // root, epoch, num_nodes
-                if let (Some(Ok(root)), Some(Ok(epoch)), Some(Ok(num_nodes))) =
-                    (row.take_opt(0), row.take_opt(1), row.take_opt(2))
-                {
-                    let azks = AsyncMySqlDatabase::build_azks(root, epoch, num_nodes);
+                // epoch, num_nodes
+                if let (Some(Ok(epoch)), Some(Ok(num_nodes))) = (row.take_opt(0), row.take_opt(1)) {
+                    let azks = AsyncMySqlDatabase::build_azks(epoch, num_nodes);
                     return Ok(DbRecord::Azks(azks));
                 }
             }
@@ -1870,13 +1925,14 @@ impl MySqlStorable for DbRecord {
                 }
             }
             StorageType::HistoryTreeNode => {
-                // `location`, `label_len`, `label_val`, `epochs`, `parent`, `node_type`
+                // `label_len`, `label_val`, `birth_epoch`, `last_epoch`, `parent_label_len`, `parent_label_val`, `node_type`
                 if let (
-                    Some(Ok(location)),
                     Some(Ok(label_len)),
                     Some(Ok(label_val)),
-                    Some(Ok(epochs)),
-                    Some(Ok(parent)),
+                    Some(Ok(birth_epoch)),
+                    Some(Ok(last_epoch)),
+                    Some(Ok(parent_label_len)),
+                    Some(Ok(parent_label_val)),
                     Some(Ok(node_type)),
                 ) = (
                     row.take_opt(0),
@@ -1885,21 +1941,18 @@ impl MySqlStorable for DbRecord {
                     row.take_opt(3),
                     row.take_opt(4),
                     row.take_opt(5),
+                    row.take_opt(6),
                 ) {
-                    let bin_vec: Vec<u8> = epochs;
-                    if let Some(decoded_epochs) = DbRecord::deserialize_epochs(&bin_vec) {
-                        let node = AsyncMySqlDatabase::build_history_tree_node(
-                            label_val,
-                            label_len,
-                            location,
-                            decoded_epochs,
-                            parent,
-                            node_type,
-                        );
-                        return Ok(DbRecord::HistoryTreeNode(node));
-                    } else {
-                        return Err(MySqlError::from("Deserialization of epochs failed"));
-                    }
+                    let node = AsyncMySqlDatabase::build_history_tree_node(
+                        label_val,
+                        label_len,
+                        birth_epoch,
+                        last_epoch,
+                        parent_label_val,
+                        parent_label_len,
+                        node_type,
+                    );
+                    return Ok(DbRecord::HistoryTreeNode(node));
                 }
             }
             StorageType::ValueState => {
@@ -1935,30 +1988,6 @@ impl MySqlStorable for DbRecord {
         let err = MySqlError::Driver(mysql_async::DriverError::FromRow { row: row.clone() });
         Err(err)
     }
-
-    fn serialize_epochs(epochs: &[u64]) -> Vec<u8> {
-        let mut results = vec![];
-        for item in epochs {
-            let bytes = (*item).to_be_bytes();
-            results.extend_from_slice(&bytes);
-        }
-        results
-    }
-
-    fn deserialize_epochs(bin: &[u8]) -> Option<Vec<u64>> {
-        if bin.len() % 8 == 0 {
-            // modulo 8 means that we have proper length byte arrays which can be decoded into u64's
-            let mut results = vec![];
-            for chunk in bin.chunks(8) {
-                let mut a: [u8; 8] = Default::default();
-                a.copy_from_slice(chunk);
-                results.push(u64::from_be_bytes(a));
-            }
-            Some(results)
-        } else {
-            None
-        }
-    }
 }
 
 trait StorageErrorWrappable {
